// wide_enrollment_task.cpp
/*
Scenario: An investment scheme is announced by a 
bank by issuing new bonds. The head office of the 
bank prepares enrolment for the scheme at each of 
its branches.  
*/ 
# include "..\..\my_essentials.hpp"

class Libor // London Interbank Offered Rate (LIBOR), which is determined by a group of banks.
{	
	double libor;
public:
	Libor(): libor(0.5) {}	
	double get_rate(){
		return libor;
	}	
};

// Description from the bank for the new bonds.
class Bond_description : public Libor {
// So Libor becomes the base class for this class
// and this class becomes the derived class 
// for Libor.
	double face_value;
	double coupon_rate;
	double coupon_frequency;
	double maturity;	
public:
	Bond_description(); 	
	void bond_details();
};

Bond_description::Bond_description() :
// Member initialization
face_value(1000),
coupon_rate((get_rate()+0.25)),
coupon_frequency(0.5),
maturity(5){
}

void Bond_description::bond_details(){
	cout << "\nBond details:\n";
	cout << "face value = " << face_value 
	 	 << " USD\n";
	cout << "coupon_rate = " << coupon_rate 
	 	 << " %\n";
	cout << "coupon_frequency = "
	 	 << coupon_frequency << " year\n";
	cout << "maturity = " << maturity 
	 	 << " years\n";
}

//----------------------------------------------

// Collection of enrolments from branch offices: 

// Note: this class provides an example for 
// illustrating the esssentail operations 
// in classes (as mentioned in section 4.2).

class collect: public Bond_description {

// So Bond_description becomes the base class for 
// this class and this class becomes the derived
// class for the Bond_description.
		
private:
	int* acc_no;
	int volume;

public:

// 1. Default constructor:

collect()
	:volume(0), acc_no(new int[0])
{
	cout << "\"default constructor\"\n"
	 	 << " constructs a pointer of value "
	 	 << acc_no << endl;
}

// 2. Constructor from argument(s):

	
explicit collect(int sz)
	:volume(sz), acc_no(new int[sz])
{
	cout
	<< "\"constructor from an argument\"\n"
	<< " constructs a pointer of value "
	<< acc_no << endl;
	for(int i =0; i != volume; ++i)
		acc_no[i] =0; 
}
		
collect(int *data, int sz)
	:volume(sz), acc_no(new int[sz])
{
cout 
<< "\"constructor from two arguments\"\n"
<< " constructs a pointer of value "
<< acc_no << endl;
	for(int i =0; i != volume; ++i)
		acc_no[i] =data[i]; 
}
	

void init_copy(const collect& document) 
/* the following function performs a Deep Copy of 
the data (as in STL containers), not a Shallow Copy 
where the address alone is copied (as in pass-by-
reference) */
{
	for(int i =0; i != document.volume;++i)
		acc_no[i] = document.acc_no[i];
}// This init_copy will be used by both the copy 
// constructor and copy assignment.

// 3. Copy constructor:

collect(const collect& document)
	:volume(document.volume), acc_no(new int[document.volume])
{
	cout 
	<< "\"copy constructor\"\n"
	<<" constructs a pointer pointing at address "
	<< acc_no << endl;
	init_copy(document);
}

// 4. Move constructor: 

collect(collect&& document)
	:volume(document.volume),
	 acc_no(document.acc_no)
{
	cout 
	<< "\n\"move constructor\"\n"
	<<" it copies the given object pointer: "
	<< document.acc_no
	<< "\n to the data pointer of this class\n";

	document.acc_no = nullptr;
	document.volume = 0;

	cout 
	<< " and release the given object pointer\n"
	<< " with a null pointer: "
	<< document.acc_no << endl;
}

// 5. Copy assignment:

collect& operator=(const collect& document)
{
	cout 
	<< "\"copy assignment\" copies\n"
	<< "from the pointer " << document.acc_no
	<< "\nto the pointer " << acc_no << endl;
	if(this==&document) return *this;
/* That is, if both have the same address then it is 
a self-assignment and hence, no further work is 
needed. If both are different and the new size is 
less than or equal to the existing size then there 
is no need for new allocation, just to copy the 
elements and assign,
size = new size and return its self-reference. */
	if(document.volume <= volume)
	{
		init_copy(document);
		volume = document.volume;
		return *this;
	}

	// so, if new size > present size

	int* temp = new int[document.volume];
	init_copy(document); /* CLASS is CLEVER 
enough to copy the document to the empty pointer 
temp rather than to the filled pointer "this". So, 
we have safely assigned the new values to the 
pointer temp. Let us now delete the document in this 
pointer.  */
 
	delete[]acc_no;

	// update this pointer with the pointer temp.
	acc_no= temp;
	volume=document.volume;

	// return a self-reference, 
	// because it is not a constructor
	return *this;
	// it is a general practice to return *this
	// in every assignment operator.
}

// 6. Move assignment:

collect& operator=(collect&& document)
{
	cout 
	<< "\n\"move assignment\"\n"
	<< " assigns pointer of this object: "
	<< acc_no << endl
	<< " to the pointer of given object: "
	<< document.acc_no << ",\n";


	if(this==&document) return *this;

	delete[]acc_no;

	acc_no= document.acc_no;
	volume=document.volume;

	document.acc_no = nullptr;
	document.volume = 0;
	
	cout
	<< " then release the pointer of the\n"
	<< " given object with a null pointer:"
	<< document.acc_no << endl;

	return *this;
}

/* 
Note: 
Here, in both the copy constructor and copy 
assignment, we copied the data members alone: acc_no 
and volume of this class collect and we did not 
copy the data members of its base class 
bond_description. This is because the data members 
of bond_description are the same for every 
acc_no. Were this not so, then we should have 
copied all of this class data members and all of its 
base class data members. 
*/

// 7. swap operation:

void swap(collect& cl1, collect& cl2)
{
	cout << "\n\"swaps data in the pointers: "
	 << cl1.acc_no << ", " << cl2.acc_no << endl;

 	collect temp = static_cast<collect &&> (cl1);
	// it calls move constructor

	cl1 = static_cast<collect &&> (cl2); 
	// it calls move assignment 

	cl2 = static_cast<collect &&> (temp); 
	// it calls move assignment

	cout << "\n\"Data in the pointers: "
	 	 << cl1.acc_no << ", " << cl2.acc_no 
		 << " are swapped" << endl;
} 

// 8. A user defined constructor:

collect(const collect& cl1, const collect& cl2)
	:volume(cl1.volume+cl2.volume),
	acc_no(new int[cl1.volume+cl2.volume])
{
	cout 
	<< "\n\"Task constructor\"\n"
    <<"constructs a pointer pointing at address "
	<< acc_no 
	<< "\nand then it fills it with the data " 
	<< "from \nthe pointers " << cl1.acc_no
	<< " and " << cl2.acc_no << "." << endl; 
	int i =0;
	for(; i != cl1.volume; ++i)
		acc_no[i] = cl1.acc_no[i];
	for(int j = 0; j != cl2.volume; ++j, ++i)
		acc_no[i] = cl2.acc_no[j];
}

friend ostream& 
	operator << (ostream & os, const collect& cl)
/* As ostream belongs to the Standard Library, which 
is out of this class, it cannot access members of 
this class unless it is a friend to this class.
So, we use the keyword friend to declare ostream 
as a friend to this class, hence it can access all 
its members. */
{
	for(int i = 0; i != cl.volume;++i)
		os 	<< '(' << i+1 << ") "
			<< cl.acc_no[i] << endl;
	return os;
}

// 9. Destructor:

~collect() 
{
	cout << "\"destructor\" destructs "
	     << acc_no << endl;
	delete[] acc_no;
}

};
// end of class "collect"
//-------------------------------------


void record()
{	
cout<< "Response to the statements:\n\n" 
 	<< "(A) Statement: \"collect record1;\":" 
 	<< endl;
collect record1;

cout << "\n(B1) \"collect record2 = record1;\": " 
 	 << endl;
collect record2 = record1;

cout << "\n(B2) \"collect record3(record1);\":" 
 	 << endl;
collect record3(record1);

cout << "\n(C) \"record3 = record2;\":" 
 	 << endl;
record3 = record2;

cout << "\n(D1) \"collect record4(5);\": " 
 	 << endl;
collect record4(5);

int* branch1 = new int[3];
branch1[0] = 123456;
branch1[1] = 654321;
branch1[2] = 123123;
	
int *branch2(new int[3]);
branch2[0] = 456456;
branch2[1] = 456123;
	
cout << "\n(D2) \"record5(branch1, 3);\": \n";
collect record5(branch1, 3);
cout << "\n(D3) \"record5(branch2, 2);\": \n";
collect record6(branch2, 2);

cout << "\nThe contents of record5:\n" << record5;
cout << "\nThe contents of record6:\n" << record6;

cout << "\n(E) An application of \"swap(record5, record6);\" \n";
record6.swap(record5, record6);

cout << "\nTherefore, after the swap:\n" 
	 << "the contents of record5:\n"
	 << record5 << endl;
cout << "the contents of record6:\n"
	 << record6;

cout << "\n(F) statement:\n\"collect record7(record5, record6);\"\n";
collect record7(record5, record6);

record7.bond_details();// Accessing member function 
// of the base class.
cout << "\nTotal list of customers enrolled:\n"
	 << record7 << endl
	 << "(G) The pointers are going out of scope, so:\n";
}

int main()
{
	record();
	just_pause();
}


/***************************************

Copyright  2013 Coding4You
N. S. Pattabi Raman 

Permission to use this software 
for academic or commercial purpose
is hereby granted to the programmers
who have purchased the book,
Concise exposure to C++.

N. S. Pattabi Raman makes no 
representations about the suitability
of this software for any purpose. 
It is provided "as is" without express 
or implied warranty. 

***************************************/
