// mutex_to_control_data_race.cpp

#include "my_essentials_and_concurrency.hpp"

// ---------------------------------------

// (a) function with no control on threads race: 

auto funct_with_no_control 
		= [](double &value, int it) {
	for(auto i = 0; i< it; ++i)
		value += 0.5;
};
// ---------------------------------------

// (b) function with mutex lock and unlock on threads race:

auto funct_with_mutex = [](double &value, int it) {	
	static mutex my_mutex;
	for(auto i = 0; i< it; ++i)
	{
		my_mutex.lock();	
		value += 0.5;
		my_mutex.unlock();
	}
};
// ---------------------------------------

// (c) function with lock_guard on threads race:

auto funct_with_lock_guard 
			= [](double &value, int it) {	
	static mutex my_mutex;
	for(auto i = 0; i< it; ++i)
	{
		lock_guard<mutex> guard(my_mutex);
		value += 0.5;
	}
};
// ---------------------------------------

class race{
	double value;
	int it;
	void try_with(function<void(double&, int)>);
	void try_all();
public:
	explicit race(int iteration)
		: it(iteration) {
			try_all();
	}
};
// ---------------------------------------

void race::try_all()
{
	cout 
	<< "\n (a) Using funct_with_no_control:\n\n";
	try_with(funct_with_no_control);

	cout << "\n (b) Using funct_with_mutex:\n\n";
	try_with(funct_with_mutex);

	cout 
	<< "\n (c) Using funct_with_lock_guard:\n\n";
	try_with(funct_with_lock_guard);
}


// ---------------------------------------

int main ()
try
{
	int it(123456);
	race rc(it);

	just_pause();
}

catch (char* e) {
    cerr << "\n Exception caught: " << e << endl;
    just_pause();
}
catch(...){
	cout	<< "\n Problem: there is an unknown " 
		<< "exception!\n";
	just_pause();
	return 2;
}

void race::try_with(function<void(double&, int)> funct)
{	
	cout << setprecision(10);

	// Serial:
	value = 0.0;
	for(auto i = 0; i < 2; ++i) {
		funct(ref(value), it);
	}
	auto stop = time_now();
	cout   << "\t\t serial: value = " 
		<< value << endl;
	//----------------------

	// Independent Parallel:
	vector<future<int>>independent_threads;
	independent_threads.reserve(2);

	value = 0.0;
	vector<double> value_separate(2,0);

	for(auto i = 0; i < 2; ++i) 				
		independent_threads.push_back(
		async(funct, ref(value_separate[i]), 
		it));

	for(auto& th: independent_threads)
		th.get();
	value = accumulate(value_separate.begin(), 
			value_separate.end(),0.0);

	stop = time_now();
	cout   << "   independent parallel: value = "
		<< value  << endl;
	//-------------------

	// Racing Parallel:
	vector<future<int>>racing_threads;
	racing_threads.reserve(2);

	value = 0.0;
	for(auto i = 0; i < 2; ++i)
		racing_threads.push_back(async(funct, 
					ref(value), it));

	for(auto& th: racing_threads)
		th.get();

	stop = time_now();
	cout   << "        racing parallel: value = " 
		<< value  << endl;
	//--------------------
}

/***************************************

Copyright  2013 Coding4You
N. S. Pattabi Raman 

Permission to use this software 
for academic or commercial purpose
is hereby granted to the programmers
who have purchased the book,
Concise exposure to C++.

N. S. Pattabi Raman makes no 
representations about the suitability
of this software for any purpose. 
It is provided "as is" without express 
or implied warranty. 

***************************************/
